"""Unit test for hhlpy."""

import unittest
import os
from wolframclient.evaluation import WolframLanguageSession

from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp.hcsp import HCSP, Function
from ss2hcsp.hcsp.parser import parse_expr_with_meta, parse_expr_with_meta, parse_hp_with_meta, \
    parse_hoare_triple_with_meta
from hhlpy.hhlpy_without_dG import CmdVerifier
from hhlpy.wolframengine_wrapper import found_wolfram, session

def runFile(self, file, 
              andR_rule=None,
              print_vcs=False, expected_vcs=None):
    # Read the file
    file = os.path.join(os.path.dirname(__file__), "../examples", file)
    file = open(file,mode='r', encoding='utf-8')
    file_contents = file.read()
    file.close()

    # Parse pre-condition, HCSP program, and post-condition
    hoare_triple = parse_hoare_triple_with_meta(file_contents)

    # Initialize the verifier
    verifier = CmdVerifier(
        pre=hoare_triple.pre, 
        hp=hoare_triple.hp,
        post=hoare_triple.post,
        functions=hoare_triple.functions)
    
    if andR_rule:
        for pos, andR in andR_rule.items():
            if isinstance(andR, str):
                andR = parse_expr_with_meta(andR)
            verifier.set_andR_rule(pos, andR)

    # Compute wp and verify
    verifier.compute_wp()

    # Optional: Print verification conditions
    if print_vcs:
        for pos, vcs in verifier.get_all_vcs().items():
            print("%s:" % str(pos))
            for vc in vcs:
                print(vc.expr, 
                "pos:", vc.pos, 
                "annot_pos:", vc.annot_pos,
                "categ:", vc.categ,
                "blabel:", str(vc.blabel),
                "comp_label:", str(vc.comp_label),
                "vc:", vc.vc,
                "pc:", vc.pc)

    # Use SMT to verify all verification conditions
    self.assertTrue(verifier.verify())

    # Optional: check the verification conditions are expected
    def is_trivial(vc):
        if isinstance(vc, expr.LogicExpr) and vc.op == "->" and vc.exprs[0] == vc.exprs[1]:
            return True
        else:
            return False

    if expected_vcs:
        for pos, vcs in expected_vcs.items():
            vcs = [parse_expr_with_meta(vc) for vc in vcs]
            actual_vcs = [vc.expr for vc in verifier.infos[pos].vcs if not is_trivial(vc.expr)]
            self.assertEqual(set(vcs), set(actual_vcs), 
            "\nExpect: {}\nActual: {}".format([str(vc) for vc in vcs],[str(vc) for vc in actual_vcs]))

class HHLPyTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if found_wolfram:
            session.start()

    @classmethod
    def tearDownClass(cls):
        if found_wolfram:
            session.terminate()

    def testParseHoareTriple(self):
        res = parse_hoare_triple_with_meta("""
            pre [x >= 0];
            x := x + 1;
            post [x >= 1];
        """)
        self.assertEqual(str(res.pre[0]), "x >= 0")
        self.assertEqual(str(res.post[0].expr), "x >= 1")
        self.assertEqual(str(res.hp), "x := x + 1;")

    def testParseHoareTriple2(self):
        res = parse_hoare_triple_with_meta("""
            function bar(x, y) = 2 * x + y;
            pre [x >= 0];
            x := x + 1;
            post [x >= 1];
        """)
        self.assertEqual(str(res.functions["bar"]), "bar(x,y) = 2 * x + y")

    def testParseHoareTriple3(self):
        res = parse_hoare_triple_with_meta("""
            function bar(x, y) = 2 * x == y;
            pre [x >= 0];
            x := x + 1;
            post [x >= 1];
        """)
        self.assertEqual(str(res.functions["bar"]), "bar(x,y) = 2 * x == y")

    def testReplaceFunction1(self):
        res = parse_expr_with_meta("bar(y)")
        funcs = {'bar': Function('bar', ['x'], 'x + 1')}

        self.assertEqual(str(expr.replace_function(res, funcs)), 'y + 1')

    def testReplaceFunction2(self):
        res = parse_expr_with_meta("bar(a, b)")
        
        funcs = {'bar': Function('bar', ['x', 'y'], '2 * x == y')}

        self.assertEqual(str(expr.replace_function(res, funcs)), '2 * a == b')

    def testReplaceFunction3(self):
        res = parse_expr_with_meta("am(x)")
        
        funcs = {'bar': Function('bar', ['x'], 'x + 1'), 'am': Function('am', ['x'], '2 * bar(x)')}

        self.assertEqual(str(expr.replace_function(res, funcs)), '2 * bar(x)')

    def testVerify1(self):
        runFile(self, file="test1.hhl")

    def testVerify1_1(self):
        runFile(self, file="test1_1.hhl")

    def testVerify2(self):
        runFile(self, file="test2.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> x + 1 >= 1", "x >= 0 -> x + 2 >= 1"]},
                  print_vcs=False)

    def testVerify2_1(self):
        runFile(self, file="test2_1.hhl",
        print_vcs=False)

    def testVerify2_2(self):
        runFile(self, file="test2_2.hhl",
                print_vcs=False)

    def testVerify2_3(self):
        runFile(self, file="test2_3.hhl",
                print_vcs=False)

    def testVerify3(self):
        runFile(self, file="test3.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> x + 1 >= 1", "x >= 0 -> x + 1 + 2 >= 3"]})

    def testVerify4_1(self):
        runFile(self, file="test4_1.hhl",
                  print_vcs=False)

    def testVerify4_2(self):
        runFile(self, file="test4_2.hhl",
                  print_vcs=False)

    def testVerify5(self):
        runFile(self, file="test5.hhl", print_vcs=False,
                  expected_vcs={((), (0,)): ["x >= 0 -> x + 1 >= 0"]})

    def testVerify5_1(self):
        runFile(self, file="test5_1.hhl",
                  print_vcs=False)

    def testVerify5_2(self):
        runFile(self, file="test5_2.hhl",
                  print_vcs=False)

    def testVerify5_3(self):
         runFile(self, file="test5_3.hhl",
                  print_vcs=False)

    def testVerify7(self):
        runFile(self, file="test7.hhl", print_vcs=False)

    # TODO: 
    # def testVerify7_1(self):
    #     # {true} <x_dot = 2 & x < 10> {true}
    #     runVerify(self, pre="true", hp="<x_dot = 2 & x < 10>", post="true", print_vcs=True)

    def testVerify7_2(self):
        runFile(self, file="test7_2.hhl",
                  print_vcs=False)

    def testVerify8(self):
        runFile(self, file="test8.hhl",
                  print_vcs=False)

    def testVerify11(self):
        runFile(self, file="test11.hhl",
                  expected_vcs={((), ()): ["x0 >= 0 -> x0 >= 0 -> x1 >= 1 -> x1 >= 1"]})

    def testVerify12(self):
        runFile(self, file="test12.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> y0 >= x + 1 -> y0 >= 1"]})

    def testVerify13(self):
        runFile(self, file="test13.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> y1 >= x + 1 ->y0 >= x + 1 + 1 -> y0 >= 2"]})

    def testVerify9(self):
        runFile(self, file="test9.hhl", print_vcs=False)

    def testVerify14_1(self):
        runFile(self, file="test14_1.hhl",
                  print_vcs=False) 

    def testVerify18(self):
        runFile(self, file="test18.hhl",
                  print_vcs=False)

    def testVerify36_1(self):
        runFile(self, file="test36_1.hhl",)

    def testVerify40(self):
        runFile(self, file="test40.hhl",)

    def testVerify42(self):
        runFile(self,file="test42.hhl",)

    def testVerify51(self):
        runFile(self, file="test51.hhl",)

    def testVerify52_1(self):
            runFile(self, file="test52_1.hhl",)

    def testVerify56_1(self):
        runFile(self, file="test56_1.hhl",)

    def testVerify65(self):
        runFile(self, file="test65.hhl",)

    def testVerify66(self):
        runFile(self, file="test66.hhl",)

    def testVerify67(self):
        runFile(self, file="test67.hhl",)

    def testVerify68(self):
        runFile(self, file="test68.hhl",)

class BasicHHLPyTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if found_wolfram:
            session.start()

    @classmethod
    def tearDownClass(cls):
        if found_wolfram:
            session.terminate()

    def testBasic1(self):
        runFile(self, file="basic1.hhl")
                  #expected_vcs={((), ()): ["x >= 0 -> x + 1 >= 1"]})

    def testBasic2(self):
        runFile(self, file="basic2.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> x + 1 + 1 >= 1", "x >= 0 -> x + 1 >= 1"]})

    def testBasic3(self):
        runFile(self, file="basic3.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> x + 1 >= 1"],
                                ((1,), (0,)): ["x >= 1 -> x + 1 >= 1"]})

    def testBasic4(self):
        runFile(self, file="basic4.hhl",
                  print_vcs=False)

    def testBasic5(self):
        runFile(self, file="basic5.hhl",
                  expected_vcs={((), ()): ["x >= 0 -> x0 >= 1 -> x0 >= 1"]})

    # TODO: Basic benchmark problem 6 is hard to translate into HCSP program.

    def testBasic7(self):
        runFile(self, file="basic7.hhl") 

    def testBasic8(self):
        runFile(self, file="basic8.hhl")

    def testBasic9(self):
        runFile(self, file="basic9.hhl")
    
    def testBasic10(self):
        runFile(self, file="basic10.hhl",
                  print_vcs=False)

    def testBasic11(self):
        runFile(self, file="basic11.hhl",)

    def testBasic12(self):
        runFile(self, file="basic12.hhl",
                  expected_vcs={((), ()): ["y >= 0 -> x >= 0 && y >= 0 -> t0 >= 0 -> t0 > 0 -> y >= 0",
                                           "y >= 0 -> x >= 0 && y >= 0 -> t0 >= 0 -> t0 > 0 -> x >= 0",
                                           "y >= 0 -> x >= 0 && y >= 0 -> t0 >= 0 -> t0 <= 0 -> x >= 0"]})

    def testBasic13(self):
        runFile(self, file="basic13.hhl",)


    def testBasic14(self):
        runFile(self, file="basic14.hhl",)

    def testBasic15(self):
        runFile(self, file="basic15.hhl")
                # expected_vcs={((), ()): ["x > 0 -> 0 < 1 -> (\exists y. x * y * y == 1)", \
                #                          "x > 0 -> 0 >= 1 -> x > 0"],
                #               ((1,), ()): ["(\exists y. x * y * y == 1) && t == 1 -> x > 0"]})

    def testBasic16(self):
        runFile(self, file="basic16.hhl",)

    def testBasic17(self):
        runFile(self, file="basic17.hhl",)
                #   expected_vcs={((), ()): ["y > 0 -> x > 0 && y > 0 -> t0 >= 0 -> t0 > 0 -> (\exists z. x * z * z == 1)",
                #                            "y > 0 -> x > 0 && y > 0 -> t0 >= 0 -> t0 <= 0 -> x > 0"],
                #                 ((1,), ()): ["y > 0 -> (\\exists z. x * z * z == 1) && t == 0 -> x > 0"]})

    def testBasic18(self):
        runFile(self, file="basic18.hhl", print_vcs=False)

    def testBasic19(self):
        runFile(self, file="basic19.hhl",)

    # TODO: Basic benchmark, problem 20. The expression is not a polynomial.

    def testBasic21(self):
        runFile(self, file="basic21.hhl",)

    def testBasic22(self):
        runFile(self, file="basic22.hhl",)

    def testBasic23(self):
        runFile(self, file="basic23.hhl",)

    def testBasic24(self):
        runFile(self, file="basic24.hhl",)

    def testBasic25(self):
        runFile(self, file="basic25.hhl",)

    def testBasic26(self):
        runFile(self, file="basic26.hhl",)

    def testBasic27(self):
        runFile(self, file="basic27.hhl",
                  expected_vcs={((),()): ["z == -2 -> x >= 1 && y == 0 -> x >= 1",
                                          "z == -2 -> x >= 1 && y == 0 -> y >= 0", 
                                          # `y == 0` comes from implicit dW
                                          "z == -2 -> x >= 1 && y == 10 && z == -2 ->\
                                          (y > 0 -> x >= 1)", 
                                          "z == -2 -> x >= 1 && y == 10 && z == -2 -> \
                                           y <= 0 -> x >= 1",
                                           "z == -2 -> x >= 1 && y == 10 && z == -2 -> \
                                           y <= 0 -> y >= 0"],
                                ((), (0,)): ["z == -2 -> y > 0 -> y >= 0"] 
                                          # This is from dI (condition implies differential of invariant)
                                }) 
                                          # `y <= 0 -> x >= 1 && y >= 0` is the dW precondition

    def testBasic28(self):
        runFile(self, file="basic28.hhl",)

    def testBasic29(self):
        runFile(self, file="basic29.hhl",)

    # TODO: Benchmark, problem 30, 32 are hard to translate into hcsp programs.

    def testBasic31(self):
        runFile(self, file="basic31.hhl",)

    def testBasic33(self):
        runFile(self, file="basic33.hhl",)

    def testBasic34(self):
        runFile(self, file="basic34.hhl",)

    def testBasic35(self):
        runFile(self, file="basic35.hhl",)

    def testBasic36(self):
        runFile(self, file="basic36.hhl",)

    def testBasic37(self):
        runFile(self, file="basic37.hhl",)

    def testBasic38(self):
        runFile(self, file="basic38.hhl",)

    def testBasic39(self):
        runFile(self, file="basic39.hhl",)

    def testBasic40(self):
        runFile(self, file="basic40.hhl",)
                #   constants={'A'})

    def testBasic41(self):
        runFile(self, file="basic41.hhl",)
                #   constants={'A', 'B'})

    def testBasic42(self):
        runFile(self, file="basic42.hhl",)
                #   constants={'A', 'B', 'S'})

    def testBasic43(self):
        runFile(self, file="basic43.hhl",)
                #   constants={'A', 'V'})

    def testBasic44(self):
        runFile(self, file="basic44.hhl",)
                #   constants={'A', 'V'})

    def testBasic45(self):
        runFile(self, file="basic45.hhl",
                #  constants={'A', 'V'}
                ) 

    def testBasic46(self):
        runFile(self, file="basic46.hhl",)

    def testBasic47(self):
        runFile(self, file="basic47.hhl",)

    def testBasic48(self):
        runFile(self, file="basic48.hhl",)

    def testBasic49(self):
        runFile(self, file="basic49.hhl",)

    def testBasic50(self):
        runFile(self, \
                  file="basic50.hhl",
                  )

    # TODO: Basic benchmark, problem 51. The ODE invariant cannot imply loop invariant.
    # def testVerify61(self):
    #     # Basic benchmark, problem 51
    #     # TODO: to implement abs, old(v)
    #     #         v >= 0 & A > 0 & B >= b & b > 0 & ep > 0 & lw > 0 & y = ly & r != 0 & dx^2 + dy^2 = 1
    #     #            & abs(y-ly) + v^2/(2*b) < lw
    #     #  -> [
    #     #       { {   ?abs(y-ly) + v^2/(2*b) + (A/b+1)*(A/2*ep^2+ep*v) < lw;
    #     #             a :=*; ?-B <= a & a <= A;
    #     #             w :=*; r :=*; ?r != 0 & w*r = v;
    #     #          ++ ?v=0; a := 0; w := 0;
    #     #          ++ a :=*; ?-B <=a & a <= -b;
    #     #         }

    #     #         c := 0;
    #     #         {
    #     #         { x' = v*dx, y' = v*dy, v' = a, dx' = -dy*w, dy' = dx*w, w'=a/r, c' = 1 & v >= 0 & c <= ep }
    #     #         @invariant(
    #     #           c>=0,
    #     #           dx^2+dy^2=1,
    #     #           (v'=0 -> v=old(v)),
    #     #           (v'=0 -> -c*v <= y - old(y) & y - old(y) <= c*v),
    #     #           (v'=a -> v=old(v)+a*c),
    #     #           (v'=a -> -c*(v-a/2*c) <= y - old(y) & y - old(y) <= c*(v-a/2*c))
    #     #         )
    #     #         }
    #     #       }*@invariant(v >= 0 & dx^2+dy^2 = 1 & r != 0 & abs(y-ly) + v^2/(2*b) < lw)
    #     #     ] abs(y-ly) < lw
    #     runVerify(self, pre="v >= 0 & A > 0 & B >= b & b > 0 & ep > 0 & lw > 0 & \
    #               y == ly & r != 0 & dx^2 + dy^2 == 1 & \
    #               y-ly < lw - v^2/(2*b) & y-ly > -lw + v^2/(2*b)",
    #               hp="( \
    #                         if y-ly < - (v^2/(2*b) + (A/b+1)*(A/2*ep^2+ep*v)) + lw &\
    #                            y-ly > (v^2/(2*b) + (A/b+1)*(A/2*ep^2+ep*v)) - lw \
    #                         then \
    #                             a := {-B <= a & a <= A}; \
    #                             r := {r != 0}; \
    #                             w := v/r \
    #                         else \
    #                             a := {-B <=a & a <= -b} \
    #                         endif \
    #                     ++  if v == 0 \
    #                         then a := 0; w := 0 \
    #                         else a := {-B <=a & a <= -b} \
    #                         endif \
    #                     ++  a := {-B <=a & a <= -b}; \
    #                     c := 0; \
    #                     v0 := v; \
    #                     y0 := y; \
    #                     <x_dot = v*dx, y_dot = v*dy, v_dot = a, dx_dot = -dy*w, dy_dot = dx*w, w_dot=a/r, c_dot = 1 & v > 0 & c < ep> \
    #                     invariant \
    #                         [c >= 0] \
    #                         [dx^2+dy^2 == 1] \
    #                         [r != 0] \
    #                         [v == v0 + a*c] \
    #                         [-c*(v-a/2*c) <= y - y0] \
    #                         [y - y0 <= c*(v-a/2*c)] \
    #                 )** invariant \
    #                         [v >= 0 & dx^2+dy^2 == 1 & r != 0 & \
    #                         y-ly < lw - v^2/(2*b) & y-ly > -lw + v^2/(2*b)]",
    #                 post="y-ly < lw - v^2/(2*b) & y-ly > -lw + v^2/(2*b)",
    #                 constants={'A', 'B', 'b', 'ep', 'lw', 'ly'},
    #                 wolfram_engine=True)

    def testBasic52(self):
        runFile(self, file="basic52.hhl",)

    def testBasic53(self):
        runFile(self, file="basic53.hhl",)

    def testBasic54(self):
        runFile(self, file="basic54.hhl",)

    def testBasic55(self):
        runFile(self, file="basic55.hhl",)
    
    # TODO: Basic benchmark, problem 56 - 60, cannot be written into hcsp program.


class NonlinearHHLPyTest(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        try:
            session.start()
        except Exception as e:
            session.start()

    @classmethod
    def tearDownClass(cls):
        session.terminate()
  
    def testNonlinear1(self):
        runFile(self, file="nonlinear1.hhl")
    
    def testNonlinear2(self):
        runFile(self, file="nonlinear2.hhl")

    def testNonlinear3(self):
        runFile(self, file="nonlinear3.hhl")

    def testNonlinear4(self):
        runFile(self, file="nonlinear4.hhl",)

    def testNonlinear5(self):
        runFile(self, file="nonlinear5.hhl")

    def testNonlinear6(self):
        runFile(self, file="nonlinear6.hhl",)

    def testNonlinear7(self):
        runFile(self, file="nonlinear7.hhl")
    
    def testNonlinear8(self):
        runFile(self, file="nonlinear8.hhl",)

    def testNonlinear9(self):
        runFile(self, file="nonlinear9.hhl",)

    # TODO: Nonlinear benchmark, problem 10.
    # Cannot proved by keymaera following tactic.
#     x^2+y^2<=1
#   ->
#   [
#     {x'=3*(-4+x^2), y'=3+x*y-y^2}
#     @invariant(
#       (-0.0013138)*x^5+0.00048141*x^4*y+0.000828*x^2*y^3-0.0016748*x*y^4+0.0008106*y^5+0.010722*x^4-0.0018729*x^3*y+0.0041383*x^2*y^2-0.013911*x*y^3+0.0085091*y^4-0.039948*x^3-0.0060006*x^2*y-0.046355*x*y^2+0.054433*y^3-0.028132*x^2+0.13217*x*y+0.10916*y^2+0.62004*x-0.88775*y-1.1161<=0&(-0.00011438)*x^4*y^2+0.00015105*x^2*y^4-0.0018063*x^5+0.0012699*x^3*y^2+0.0014498*x^2*y^3-0.0014334*x*y^4+0.0013001*y^5+0.017567*x^4+0.0050023*x^3*y-0.0016674*x^2*y^2-0.015315*x*y^3+0.01038*y^4-0.072259*x^3-0.035874*x^2*y-0.050558*x*y^2+0.058708*y^3-0.05097*x^2+0.042626*x*y+0.19257*y^2+1.3148*x+0.014613*y-1.2585<=0,
#       x^2+y^2-16<=0)
#   ] !(x < -4 | y < -4 | x>4 | y>4) 
    
    def testNonlinear11(self):
        runFile(self, file="nonlinear11.hhl",)

    def testNonlinear12(self):
        runFile(self, file="nonlinear12.hhl")

    def testNonlinear13(self):
        runFile(self, file="nonlinear13.hhl")

    def testNonlinear14(self):
        runFile(self, file="nonlinear14.hhl")

    def testNonlinear15(self):
        runFile(self, file="nonlinear15.hhl")

    def testNonlinear16(self):
        runFile(self, file="nonlinear16.hhl")

    def testNonlinear17(self):
        runFile(self, file="nonlinear17.hhl")

    def testNonlinear18(self):
        runFile(self, file="nonlinear18.hhl")

    # TODO: Nonlinear benchmark, problem 19, 20. The ODE rule is not clear.

    def testNonlinear21(self):
        runFile(self, file="nonlinear21.hhl",)

    def testNonlinear22(self):
        runFile(self, file="nonlinear22.hhl")

    def testNonlinear23(self):
        runFile(self, file="nonlinear23.hhl")
                  
    # TODO: Nonlinear benchmark, problem 24. ODE rule not clear.

    def testNonlinear25(self):
        runFile(self, file="nonlinear25.hhl")

    def testNonlinear26(self):
        runFile(self, file="nonlinear26.hhl")

    def testNonlinear27(self):
        runFile(self, file="nonlinear27.hhl")

    def testNonlinear28(self):
        runFile(self, file="nonlinear28.hhl")

    def testNonlinear29(self):
        runFile(self, file="nonlinear29.hhl")

    def testNonlinear30(self):
        runFile(self, file="nonlinear30.hhl")

    def testNonlinear31(self):
        runFile(self, file="nonlinear31.hhl")

    def testNonlinear32(self):
        runFile(self, file="nonlinear32.hhl")

    def testNonlinear33(self):
        runFile(self, file="nonlinear33.hhl")

    def testNonlinear34(self):
        runFile(self, file="nonlinear34.hhl")

    def testNonlinear35(self):
        runFile(self, file="nonlinear35.hhl")

    def testNonlinear36(self):
        runFile(self, file="nonlinear36.hhl")

    def testNonlinear37(self):
        runFile(self, file="nonlinear37.hhl")

    def testNonlinear38(self):
        runFile(self, file="nonlinear38.hhl")

    def testNonlinear39(self):
        runFile(self, file="nonlinear39.hhl")

    def testNonlinear40(self):
        runFile(self, file="nonlinear40.hhl")

    def testNonlinear41(self):
        runFile(self, file="nonlinear41.hhl")

    def testNonlinear42(self):
        runFile(self, file="nonlinear42.hhl")

    def testNonlinear43(self):
        runFile(self, file="nonlinear43.hhl")

    def testNonlinear44(self):
        runFile(self, file="nonlinear44.hhl")

    def testNonlinear45(self):
        runFile(self, file="nonlinear45.hhl")

    def testNonlinear46(self):
        runFile(self, file="nonlinear46.hhl")

    def testNonlinear47(self):
        runFile(self, file="nonlinear47.hhl")

    def testNonlinear48(self):
        runFile(self, file="nonlinear48.hhl")

    def testNonlinear49(self):
        runFile(self, file="nonlinear49.hhl")

    def testNonlinear50(self):
        runFile(self, file="nonlinear50.hhl")

    def testNonlinear51(self):
        runFile(self, file="nonlinear51.hhl")

    def testNonlinear52(self):
        runFile(self, file="nonlinear52.hhl")

    def testNonlinear53(self):
        runFile(self, file="nonlinear53.hhl")

    # TODO: Problem 54, invariant unknown

    def testNonlinear55(self):
        runFile(self, file="nonlinear55.hhl")

    # TODO: nonlinear problem 56, ODE rule is not clear. Not proved by keymaerax yet.
    # def testNonlinear56(self):
    #     # Nonlinear benchmark, problem 56
    #     # {-3<=x & x<=-2 & -3<=y & y<=-2}
    #     #      t := 0;
    #     #     <x_dot = 4 + 21*x - 7*y + 42*x^2 - 24*x*y + 4*y^2 + 27*x^3 - 9*x^2*y + 6*x^4,
    #     #      y_dot = 12 + 71*x -21*y + 150*x^2 -80*x*y + 12*y^2 + 99*x^3 - 39*x^2*y + 2*x*y^2 + 18*x^4, 
    #     #      t_dot = 1 & t < 10>
    #     #     @invariant(3+x>=0, 3+y>=0, 4+3*x*(1+x)*(7+x*(7+2*x))+4*y^2>(7+3*x*(8+3*x))*y, 1+x*(3+x)>y, 1+3*x*(2+x)>2*y, 2+3*x < y, x < 0, y < 0)
    #     # {!(x >= 0 | y >= 0)}
    #     runVerify(self, pre="-3<=x & x<=-2 & -3<=y & y<=-2",
    #               hp="t := 0; \
    #                 <x_dot = 4 + 21*x - 7*y + 42*x^2 - 24*x*y + 4*y^2 + 27*x^3 - 9*x^2*y + 6*x^4, \
    #                 y_dot = 12 + 71*x -21*y + 150*x^2 -80*x*y + 12*y^2 + 99*x^3 - 39*x^2*y + 2*x*y^2 + 18*x^4, \
    #                 t_dot = 1 & t < 10>",
    #               post="!(x >= 0 | y >= 0)",
    #               diff_cuts={((1,), ()): ["3+x >= 0", "3+y>=0", "4+3*x*(1+x)*(7+x*(7+2*x))+4*y^2>(7+3*x*(8+3*x))*y", "..."]},
    #               darboux_rule={((1,), (0,)): "true",
    #                             ((1,), (1,)): "true",
    #                             ((1,), (2,)): "true"},
    #               wolfram_engine=True)

    # TODO: Problem 57. Invariant not known.

    def testNonlinear58(self):
        runFile(self, file="nonlinear58.hhl")

    def testNonlinear59(self):
        runFile(self, file="nonlinear59.hhl")

    def testNonlinear60(self):
        runFile(self, file="nonlinear60.hhl")

    # TODO: Nonlinear problem 61. Invariants unknown.

    def testNonlinear62(self):
        runFile(self, file="nonlinear62.hhl")

    def testNonlinear63(self):
        runFile(self, file="nonlinear63.hhl")

    def testNonlinear64(self):
        runFile(self, file="nonlinear64.hhl")

    # TODO: Nonlinear 65, 66, 67. Invariants unknown.

    def testNonlinear68(self):
        runFile(self, file="nonlinear68.hhl")

    def testNonlinear69(self):
        runFile(self, file="nonlinear69.hhl")

    # TODO: Nonlinear 70. Invariants unknown.

    def testNonlinear71(self):
        runFile(self, file="nonlinear71.hhl")

    def testNonlinear72(self):
        runFile(self, file="nonlinear72.hhl")

    def testNonlinear73(self):
        runFile(self, file="nonlinear73.hhl")

    def testNonlinear74(self):
        runFile(self, file="nonlinear74.hhl")

    def testNonlinear75(self):
        runFile(self, file="nonlinear75.hhl")

    def testNonlinear76(self):
        runFile(self, file="nonlinear76.hhl")

    def testNonlinear77(self):
        runFile(self, file="nonlinear77.hhl")

    # TODO: Nonlinear problem 78 is the same with problem 77.
    def testNonlinear78(self):
        runFile(self, file="nonlinear78.hhl")

    def testNonlinear79(self):
        runFile(self, file="nonlinear79.hhl")

    def testNonlinear80(self):
        runFile(self, file="nonlinear80.hhl")

    def testNonlinear81(self):
        runFile(self, file="nonlinear81.hhl")

    # TODO: Nonlinear problem 82, Proving invariant stuck.

    def testNonlinear83(self):
        runFile(self, file="nonlinear83.hhl")

    # TODO: Nonlinear problem 84, invariant not implied by precondition

    def testNonlinear85(self):
        runFile(self, file="nonlinear85.hhl")

    # TODO: Program 86 should also be proved by dbx automatically. 
    # But now it can only proved by dI or by offering the parameter g.
    def testNonlinear86(self):
        runFile(self, file="nonlinear86.hhl")

    def testNonlinear87(self):
        runFile(self, file="nonlinear87.hhl")

    def testNonlinear88(self):
        runFile(self, file="nonlinear88.hhl")
    
    def testNonlinear89(self):
        runFile(self, file="nonlinear89.hhl")

    def testNonlinear90(self):
        runFile(self, file="nonlinear90.hhl")

    # TODO: Nonlinear benchmark, problem 91. Invariants unknown.

    # TODO: Nonlinear benchmark, problem 92, 93, 94, 95. ODE rule unknown.

    def testNonlinear96(self):
        runFile(self, file="nonlinear96.hhl")

    def testNonlinear97(self):
        runFile(self, file="nonlinear97.hhl")

    # TODO:Nonlinear problem 98. How to define a function.
    def testNonlinear98(self):
        runFile(self, file="nonlinear98.hhl")

    # TODO:Nonlinear problem 99. No tactic offered.

    # TODO:Nonlinear problem 100, result is false.
    # def testNonlinear100(self):
    #     # Nonlinear benchmark, problem 100
    #     # {x1 == 1/20 & x2 == 1/32 & sint==0 & cost==1} /* t==0 initially */
    #     #    t := 0;
    #     #   <x1_dot=x2, x2_dot=-x2-(2+sint)*x1, sint_dot=cost, cost_dot=-sint,
    #     #    t_dot = 1 & t < 10>
    #     #   @invariant(cost^2+sint^2<=100000/99999, cost^2+sint^2>=1)
    #     # {98*x1^2 + 24*x1*x2 + 24*x2*x1 + 55*x2^2 < 1}
    #     runVerify(self, pre="x1 == 1/20 & x2 == 1/32 & sint==0 & cost==1",
    #               hp="t := 0; \
    #                  <x1_dot=x2, x2_dot=-x2-(2+sint)*x1, sint_dot=cost, cost_dot=-sint, \
    #                  t_dot = 1 & t < 10>",
    #               post="98*x1^2 + 24*x1*x2 + 24*x2*x1 + 55*x2^2 < 1",
    #               diff_cuts={((1,), ()): ["cost^2+sint^2<=100000/99999", 
    #                                        "cost^2+sint^2>=1",
    #                                        "98*x1^2 + 24*x1*x2 + 24*x2*x1 + 55*x2^2 < 1"]},
    #               darboux_rule={((1,), (2,)): "true"},
    #               wolfram_engine=True)

    # TODO: Nonlinear problem 101. No tactic. Post not implied by inv.

    def testNonlinear102(self):
        runFile(self, file="nonlinear102.hhl")

    def testNonlinear103(self):
        runFile(self, file="nonlinear103.hhl")

    def testNonlinear104(self):
        runFile(self, file="nonlinear104.hhl")

    # TODO: Nonlinear problem 105, 106, 107. No tactic offered.

    # TODO: Nonlinear problem 108. Auto rule unknown. Too slow to give a result.

    def testNonlinear109(self):
        runFile(self, file="nonlinear109.hhl")

    # TODO: Nonlinear problem 110, 111. Definitions and no tactic.

    def testNonlinear112(self):
        runFile(self, file="nonlinear112.hhl")

    # TODO: Nonlinear problem 113. Definitions of constants value and no tactic.

    # TODO: Nonlinear problem 114, 115. No tactic and even inv->post is too slow to verify.

    # TODO: 
    # When not sure that ODE is executed or not, set invariant as true, 
    # because both cases should hold: 
    # Boundary of D -> Q and (P & !D) -> Q
    def testNonlinear116(self):
        runFile(self, file="nonlinear116.hhl")
                # #   diff_cuts={((), ()): ["x1+x2+x5 >= 0", "!(x2 < 0 & x1 < x2 & x5 < x2)"]},
                # #   darboux_rule={((), (0,)): "true"},
                #   diff_weakening_rule={((), ()): "true"})
        # I think this problem is different in hhl.

    # TODO: Nonlinear problem 117. No invariants offered.

    # I took out this test because it was too slow -- Alex
    # def testNonlinear118(self):
    #     runFile(self, file="nonlinear118.hhl",
    #               wolfram_engine=True)

    def testNonlinear119(self):
        runFile(self, file="nonlinear119.hhl")

    def testNonlinear120(self):
        runFile(self, file="nonlinear120.hhl")
    
    def testNonlinear121(self):
        runFile(self, file="nonlinear121.hhl")

    # TODO: Nonlinear problem 122, 123, 124. No invariants.

    # I took out this test because it was too slow -- Alex
    # def testNonlinear125(self):
    #     runFile(self, file="nonlinear125.hhl")

    def testNonlinear126(self):
        runFile(self, file="nonlinear126.hhl")

    def testNonlinear127(self):
        runFile(self, file="nonlinear127.hhl")

    # TODO: Nonlinear 128, 129, 130, 131, 132. No invariants.
    # 128: Automation in keymaera doesn't work.

    def testNonlinear133(self):
        runFile(self, file="nonlinear133.hhl")

    # TODO: Nonlinear 134. No tactic.

    def testNonlinear135(self):
        runFile(self, file="nonlinear135.hhl")

    # TODO: Nonlinear 136, 137, 138, 139, 140, 141. No tactics.

class SSHHLPyTest(unittest.TestCase):

    def testBouncing(self):
        runFile(self, file="simulink/sf_bouncing.hhl")

    def testSawtooth1(self):
        runFile(self, file="simulink/sf_sawtooth1.hhl")

    def testSawtooth2(self):
        runFile(self, file="simulink/sf_sawtooth2.hhl")

    def testDelay(self):
        runFile(self, file="simulink/sl_delay.hhl")


if __name__ == "__main__":
    unittest.main()




"""Wrapper for Z3 prover."""

from decimal import Decimal
from fractions import Fraction
import z3

from ss2hcsp.hcsp import expr

def convert(e, functions):
    """Conversion from expression to z3 term."""
    if isinstance(e, expr.AVar):
        return z3.Real(e.name)
    elif isinstance(e, expr.AConst):
        if isinstance(e.value, int):
            return z3.RealVal(e.value)
        elif isinstance(e.value, (Decimal, Fraction)):
            return z3.RealVal(str(e.value))
        else:
            print(e.value, type(e.value))
            raise NotImplementedError
    elif isinstance(e, expr.BConst):
        return z3.BoolVal(e.value)
    elif isinstance(e, expr.FunExpr):
        if len(e.exprs) == 0:  # actually a constant
            return z3.Real(e.fun_name)
        else:
            f = functions[e.fun_name]
            return f(*[convert(e, functions) for e in e.exprs])
    elif isinstance(e, expr.LogicExpr):
        if e.op == '->':
            return z3.Implies(convert(e.exprs[0], functions), convert(e.exprs[1], functions))
        elif e.op == '&&':
            return z3.And(convert(e.exprs[0], functions), convert(e.exprs[1], functions))
        elif e.op == '||':
            return z3.Or(convert(e.exprs[0], functions), convert(e.exprs[1], functions))
        elif e.op == '<->':
            return convert(e.exprs[0], functions) == convert(e.exprs[1], functions)
        elif e.op == '!':
            return z3.Not(convert(e.exprs[0], functions))
        else:
            raise TypeError
    elif isinstance(e, expr.RelExpr):
        if e.op == '<':
            return convert(e.expr1, functions) < convert(e.expr2, functions)
        elif e.op == '<=':
            return convert(e.expr1, functions) <= convert(e.expr2, functions)
        elif e.op == '>':
            return convert(e.expr1, functions) > convert(e.expr2, functions)
        elif e.op == '>=':
            return convert(e.expr1, functions) >= convert(e.expr2, functions)
        elif e.op == '==':
            return convert(e.expr1, functions) == convert(e.expr2, functions)
        elif e.op == '!=':
            return z3.Not(convert(e.expr1, functions) == convert(e.expr2, functions))
    elif isinstance(e, expr.OpExpr):
        if len(e.exprs) == 1:
            return -convert(e.exprs[0], functions)
        elif e.op == '+':
            return convert(e.exprs[0], functions) + convert(e.exprs[1], functions)
        elif e.op == '-':
            return convert(e.exprs[0], functions) - convert(e.exprs[1], functions)
        elif e.op == '*':
            return convert(e.exprs[0], functions) * convert(e.exprs[1], functions)
        elif e.op == '/':
            return convert(e.exprs[0], functions) / convert(e.exprs[1], functions)
        elif e.op == '^':
            return convert(e.exprs[0], functions) ** convert(e.exprs[1], functions)
        else:
            raise NotImplementedError
    
    elif isinstance(e, expr.ExistsExpr):
        if isinstance(e.vars, tuple):
            return z3.Exists(list(convert(var, functions) for var in e.vars), convert(e.expr, functions))
        else:
            return z3.Exists(convert(e.vars, functions), convert(e.expr, functions))
    elif isinstance(e, expr.ForAllExpr):
        if isinstance(e.vars, tuple):
            return z3.ForAll(list(convert(var, functions) for var in e.vars), convert(e.expr, functions))
        else:
            return z3.ForAll(convert(e.vars, functions), convert(e.expr, functions))
    elif isinstance(e, expr.IfExpr):
        return z3.If(convert(e.cond, functions), convert(e.expr1, functions), convert(e.expr2, functions))
    else:
        print(e, type(e))
        raise NotImplementedError

def convertFunDecl(funDecl, z3functions):
    convertedBody = convert(funDecl.expr, z3functions)
    f = z3.Function(funDecl.name, *[z3.RealSort() for v in funDecl.vars], convertedBody.sort())
    vars = [z3.Real(v) for v in funDecl.vars]
    return (f, z3.ForAll(vars, f(vars) == convertedBody))

def z3_prove(e, functions=dict()):
    """Attempt to prove the given condition."""
    s = z3.Solver()
    z3functions = {}
    for f in functions:
        z3fun, faxiom = convertFunDecl(functions[f], z3functions)
        s.add(faxiom)
        z3functions[f] = z3fun
    s.add(z3.Not(convert(e, z3functions)))
    if str(s.check()) == 'unsat':
        return True
    else:
        return False



from ss2hcsp.sl.SubSystems.subsystem import Triggered_Subsystem
from ss2hcsp.sf.sf_state import AND_State, OR_State


class SF_Chart(Triggered_Subsystem):
    def __init__(self, name, state, data, num_src, num_dest, st):
        super(SF_Chart, self).__init__(name, num_src, num_dest, None)

        self.type = "stateflow"
        self.st = st

        self.input_events = list()  # [(trigger_type, event)]
        self.exec_name = ""

        self.src_lines = [[] for _ in range(self.num_src)]  # [[]] * self.num_src
        self.dest_lines = [None] * self.num_dest

        assert isinstance(state, (AND_State, OR_State))
        self.diagram = state
        self.diagram.chart = self

        self.all_states = state.get_all_descendants() # dict
        assert self.diagram.ssid not in self.all_states
        self.all_states[self.diagram.ssid] = self.diagram

        self.data = data
        self.all_vars = sorted(data.keys())

        self.port_to_in_var = dict()
        self.port_to_out_var = dict()

    def __str__(self):
        input_events_str = ""
        if self.input_events:
            for trigger_type, event in self.input_events:
                input_events_str += "Input event %s, %s\n" % (trigger_type, event)
        return "Chart(%s):\n%s%s" % (self.name, input_events_str, str(self.diagram))



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, RelExpr


class Relation(SL_Block):
    """Relation of two dest lines."""
    def __init__(self, name, relation, *, st=-1):
        super(Relation, self).__init__()
        self.name = name
        self.type = "relation"
        self.is_continuous = (st == 0)
        self.num_src = 1
        self.num_dest = 2
        self.src_lines = [[]]
        self.dest_lines = [None] * self.num_dest

        assert relation in ["<", ">", "==", "!=", ">=", "<="]
        self.relation = relation

        assert isinstance(st, (int, float))
        self.st = st

    def __str__(self):
        return "%s: Relation[in = %s, out = %s, st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)

    def get_var_map(self):
        in_vars = [AVar(line.name) for line in self.dest_lines]
        cond0 = RelExpr(self.relation, in_vars[0], in_vars[1])
        expr0 = AConst(1)
        cond1 = cond0.neg()
        expr1 = AConst(0)
        out_var = self.src_lines[0][0].name
        return {out_var: [(cond0, expr0), (cond1, expr1)]}



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, RelExpr


class Reference(SL_Block):
    def __init__(self, name, relop, st=-1):
        super(Reference, self).__init__()
        self.name = name
        self.type = "reference"
        self.is_continuous = (st == 0)
        self.num_src = 1
        self.num_dest = 1
        self.src_lines = [[]]
        self.dest_lines = [None]

        assert relop in ["<=", "<", ">", ">=", "==", "~="]
        self.relop = relop

        assert isinstance(st, (float, int))
        self.st = st

    def __str__(self):
        return "%s: Reference[in = %s, out = %s, st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)

    def get_var_map(self):
        in_var = AVar(self.dest_lines[0].name)
        cond0 = RelExpr(self.relop, in_var, AConst(0))
        expr0 = AConst(1)
        cond1 = cond0.neg()
        expr1 = AConst(0)
        out_var = self.src_lines[0][0].name
        return {out_var: [(cond0, expr0), (cond1, expr1)]}



"""Translation from diagrams to HCSP processes."""
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import *
from ss2hcsp.sl.sl_block import SL_Block, get_gcd
from ss2hcsp.sl.Continuous.signalBuilder import SignalBuilder
from itertools import product
import operator
from ss2hcsp.hcsp.parser import expr_parser, hp_parser
from ss2hcsp.sl.mux.mux import Mux
from ss2hcsp.hcsp.module import HCSPModule
from ss2hcsp.hcsp import hcsp
from ss2hcsp.sf import sf_convert
from ss2hcsp.hcsp.optimize import full_optimize_module
from ss2hcsp.sl.sl_diagram import SL_Diagram


def translate_continuous(diagram):
    # Get block dictionary
    block_dict = {block.name: block for block in diagram}
    # Get input and output channels
    in_channels, out_channels = [], []
    for block in block_dict.values():
        for line in block.dest_lines:
            if line.src not in block_dict:
                in_channels.append(hp.InputChannel(ch_name=line.ch_name, var_name=line.name))
        for lines in block.src_lines:
            for line in lines:
                if line.dest not in block_dict:
                    out_channels.append(hp.OutputChannel(ch_name=line.ch_name, expr=AVar(line.name)))
    in_channels.sort(key=operator.attrgetter("ch_name"))
    out_channels.sort(key=operator.attrgetter("ch_name"))

    # Get initial process and ODEs
    init_hps = []
    ode_eqs = []
    for block in block_dict.values():
        if block.type == "integrator":
            out_vars = block.get_output_vars()
            assert isinstance(out_vars, set) and len(out_vars) == 1
            out_var = out_vars.pop()
            init_hps.append(hp.Assign(out_var, AConst(block.init_value)))
            in_vars = block.get_input_vars()
            assert isinstance(in_vars, set) and len(in_vars) == 1
            in_var = in_vars.pop()
            ode_eqs.append((out_var, AVar(in_var)))
        # elif block.type == "unit_delay":
        #     out_vars = block.get_output_vars()
        #     assert isinstance(out_vars, set) and len(out_vars) == 1
        #     out_var = out_vars.pop()
        #     ode_eqs.append((out_var, AConst(0)))
    # Delete integrator blocks
    integator_names = [name for name, block in block_dict.items() if block.type == "integrator"]
    for name in integator_names:
        del block_dict[name]

    # Constant blocks are for substitution
    var_substitute = Conditional_Inst()  # an object for variable substitution
    for block in block_dict.values():
        if block.type == "constant":
            var_map = block.get_var_map()
            assert len(var_map) == 1
            out_var, cond_inst = var_map.popitem()
            var_substitute.add(out_var, cond_inst)
    # Delete constant blocks
    constant_names = [name for name, block in block_dict.items() if block.type == "constant"]
    for name in constant_names:
        del block_dict[name]

    # Variable substitution
    while block_dict:
        block_pool = dict()
        for name, block in block_dict.items():
            src_blocks = block.get_src_blocks()
            assert isinstance(src_blocks, set)
            if src_blocks.isdisjoint(set(block_dict.keys())):
                assert name not in block_pool
                block_pool[name] = block
        assert block_pool
        for block in block_pool.values():
            assert len(block.get_var_map()) == 1  # for current version
            for out_var, cond_inst in block.get_var_map().items():
                var_substitute.add(out_var, cond_inst)
        # Delete blocks in block_pool from block_dict
        for name in block_pool.keys():
            del block_dict[name]

    # Delete useless variables from var_substitute
    useful_vars = [in_var.name for _, in_var in ode_eqs]  # in_vars of ODEs are useful
    # Expressions (variables) on output channels are useful
    useful_vars.extend(out_ch.expr.name for out_ch in out_channels)
    useless_vars = set(var_substitute.data.keys()) - set(useful_vars)
    for var in useless_vars:
        del var_substitute.data[var]

    # Get composite condition (evolution domain for ODEs) from var_substitute
    var_list = list(var_substitute.data.keys())
    cond_inst_list = list(var_substitute.data.values())
    assert all(var_substitute.data[var_list[i]] == cond_inst_list[i] for i in range(len(var_substitute.data)))
    cond_to_var_to_expr = dict()  # var can be substituted by expr if cond is saftisfied
    for composition in product(*cond_inst_list):
        # composition is in form of [(cond0, expr0), ..., (condn, exprn)]
        conditions = set(cond for cond, _ in composition)
        if var_substitute.conflicting(conditions):
            continue
        comp_cond = conj(*conditions)  # composite condition
        assert comp_cond not in cond_to_var_to_expr
        cond_to_var_to_expr[comp_cond] = dict()
        assert len(var_list) == len(composition)
        for var, (_, expr) in zip(var_list, composition):
            cond_to_var_to_expr[comp_cond][var] = expr

    # Get ODEs by cond_to_var_expr
    ode_hps = []
    for cond, var_to_expr in cond_to_var_to_expr.items():
        # Update ode_eqs by var_to_expr
        new_ode_eqs = []
        for out_var, in_var in ode_eqs:
            assert isinstance(out_var, str) and isinstance(in_var, AVar)
            if in_var.name in var_to_expr:
                in_var = var_to_expr[in_var.name]
                assert isinstance(in_var, Expr)
            new_ode_eqs.append((out_var, in_var))
        # Update expressions on output channels by var_to_expr
        new_out_channels = []
        for out_ch in out_channels:
            assert isinstance(out_ch.expr, AVar)
            if out_ch.expr.name in var_to_expr:
                new_expr = var_to_expr[out_ch.expr.name]
                assert isinstance(new_expr, Expr)
                new_out_channels.append(hp.OutputChannel(ch_name=out_ch.ch_name, expr=new_expr))
            else:
                new_out_channels.append(hp.OutputChannel(ch_name=out_ch.ch_name, expr=out_ch.expr))
        io_comms = [(io_ch, hp.Skip()) for io_ch in in_channels + new_out_channels]
        if io_comms:
            ode_hps.append(hp.ODE_Comm(eqs=new_ode_eqs, constraint=cond, io_comms=io_comms))
        else:
            ode_hps.append(hp.ODE(eqs=new_ode_eqs, constraint=cond))

    # # Initialise input variables (Xiong: It seems not necessary, because initially,
    # # input variables should receive initial values along input channels
    # # from the environment (discrete processes) )
    # initialised_vars = [init_hp.var_name for init_hp in init_hps]
    # assert len(initialised_vars) == len(set(initialised_vars))  # no repeated initialised varaibles
    # # Xiong: Initially, the continous process should send the initial values of all output variables,
    # # and then receive the intial values of all input variables.
    out_comms = []
    in_comms = []
    for ode_hp in ode_hps:
        if isinstance(ode_hp, hp.ODE_Comm):
            for io_comm in ode_hp.io_comms:
                if isinstance(io_comm[0], hp.OutputChannel):
                    out_comms.append((io_comm[0], hp_parser.parse("num := num - 1;")))
                elif isinstance(io_comm[0], hp.InputChannel):
                    in_comms.append((io_comm[0], hp_parser.parse("num := num - 1;")))
                else:
                    raise RuntimeError("It must be a channel operation!")

    send_out_vars = hp.Skip()  # no output channel operations
    if len(out_comms) == 1:
        send_out_vars = out_comms[0][0]
    elif len(out_comms) >= 2:
        send_out_vars = hp.Sequence(hp_parser.parse("num := %s;" % len(out_comms)),
                                    hp.Loop(hp=hp.SelectComm(*out_comms), constraint=expr_parser.parse("num > 0")))

    receive_in_vars = hp.Skip()  # no input channel operations
    if len(in_comms) == 1:
        receive_in_vars = in_comms[0][0]
    elif len(in_comms) >= 2:
        receive_in_vars = hp.Sequence(hp_parser.parse("num := %s;" % len(in_comms)),
                                      hp.Loop(hp=hp.SelectComm(*in_comms), constraint=expr_parser.parse("num > 0")))

    init_hps.extend([send_out_vars, receive_in_vars])

    # ch_hp = io_comm[0]
    # if isinstance(ch_hp, hp.InputChannel):
    #     var_name = ch_hp.var_name
    #     if var_name not in initialised_vars:
    #         # updated by lqq
    #         init_hps.append(hp.Assign(var_name, AConst(1)))
    #         initialised_vars.append(var_name)
    assert init_hps
    init_ode = hp.Sequence(*init_hps)

    assert ode_hps
    # cond_ode_hps = []
    # for ode_hp in ode_hps:
    #     assert isinstance(ode_hp, (hp.ODE, hp.ODE_Comm))
    #     if ode_hp.constraint == true_expr:
    #         cond_ode_hps.append(ode_hp)
    #     else:
    #         cond_ode_hps.append(hp.Condition(cond=ode_hp.constraint, hp=ode_hp))
    # result_hp = hp.Sequence(init_ode, hp.Loop(cond_ode_hps[0])) if len(cond_ode_hps) == 1 \
    #     else hp.Sequence(init_ode, hp.Loop(hp.Sequence(*cond_ode_hps)))
    result_hp = hp.Sequence(init_ode, hp.Loop(ode_hps[0])) if len(ode_hps) == 1\
        else hp.Sequence(init_ode, hp.Loop(hp.Sequence(*ode_hps)))
    return result_hp


def translate_discrete(diagram):
    def get_block_hp(_block):  # Get the hcsp of a block from its var_map
        if _block.type == "unit_delay":
            _in_var = _block.dest_lines[0].name
            _out_var = _block.src_lines[0][0].name
            delay_out_var = "delay_" + _out_var
            return hp.Sequence(hp.Assign(_out_var, AVar(delay_out_var)),
                               hp.Assign(delay_out_var, AVar(_in_var)))

        _var_map = _block.get_var_map()
        processes = []
        for _out_var, cond_expr_list in _var_map.items():
            assert all(isinstance(_cond, Expr) and isinstance(_expr, (Expr, Expr))
                       for _cond, _expr in cond_expr_list) and cond_expr_list
            if len(cond_expr_list) == 1:
                assert cond_expr_list[0][0] == true_expr
                _expr = cond_expr_list[0][1]
                processes.append(hp.Assign(_out_var, _expr))
            elif len(cond_expr_list) >= 2:
                cond_hp_list = [(_cond, hp.Assign(_out_var, _expr)) for _cond, _expr in cond_expr_list]
                if_hps = cond_hp_list[:-1]
                else_hp = cond_hp_list[-1][1]
                processes.append(hp.ITE(if_hps, else_hp))
        assert processes
        if len(processes) == 1:
            return processes[0]
        else:
            return hp.Sequence(*processes)

    # Topological sorting for blocks in the discrete subdiagram
    def topo_sort():
        while block_dict:
            block_pool = dict()
            for name, block in block_dict.items():
                src_blocks = block.get_src_blocks()
                assert isinstance(src_blocks, set)
                if src_blocks.isdisjoint(set(block_dict.keys())):
                    assert name not in block_pool
                    block_pool[name] = block
            if not block_pool:
                break_loop()
                break
            # Classify blocks in block_pool by Sample Time
            st_to_hps = dict()
            st_to_in_chs = dict()
            st_to_out_chs = dict()
            for block in block_pool.values():
                # Get the hcsp of the block
                if block.st not in st_to_hps:
                    assert (block.st not in st_to_in_chs) and (block.st not in st_to_out_chs)
                    st_to_hps[block.st] = []
                    st_to_in_chs[block.st] = []
                    st_to_out_chs[block.st] = []
                st_to_hps[block.st].append(get_block_hp(block))
                # Get the input channels of the block
                for line in block.dest_lines:
                    if line.src not in all_blocks:
                        st_to_in_chs[block.st].append(hp.InputChannel(ch_name=line.ch_name, var_name=line.name))
                st_to_in_chs[block.st].sort(key=operator.attrgetter("ch_name"))
                for lines in block.src_lines:
                    for line in lines:
                        if line.dest not in all_blocks:
                            st_to_out_chs[block.st].append(hp.OutputChannel(ch_name=line.ch_name, expr=AVar(line.name)))
                st_to_out_chs[block.st].sort(key=operator.attrgetter("ch_name"))
            # Get each process wrt. Sample Time
            for st in st_to_hps.keys():
                # The condition of time is in form of t%st == 0
                cond_time = RelExpr("==", OpExpr("%", AVar("t"), AConst(st)), AConst(0))
                # The process is in form of in_chs?;hcsp;out_chs!
                assert st_to_hps[st]
                st_processes = st_to_in_chs[st] + st_to_hps[st] + st_to_out_chs[st]
                st_process = st_processes[0] if len(st_processes) == 1 else hp.Sequence(*st_processes)
                main_processes.append(hp.Condition(cond_time, st_process))
            # Delete blocks in block_pool from block_dict
            for name in block_pool.keys():
                del block_dict[name]

    # Break the loop in the remaining diagram
    # At present, we assume the loop is caused by unit_delay blocks
    def break_loop():
        # Find a head: the block whose source blocks in the current diagram are all unit_delay blocks
        head_block = SL_Block()
        for name, block in block_dict.items():
            src_blocks = block.get_src_blocks()
            assert isinstance(src_blocks, set)
            if all(block_dict[src_block].type == "unit_delay"
                   for src_block in src_blocks & set(block_dict.keys())):
                head_block = block
                break

        # Translate the head block
        # First, get all the input channels of the head block
        in_chs_head = []
        for line in head_block.dest_lines:
            # The head block can only read the delayed (old) value of the unit_delay blocks
            if line.src in block_dict:
                assert block_dict[line.src].type == "unit_delay"
                line.name = "delay_" + line.name
            elif line.src not in all_blocks:
                in_chs_head.append(hp.InputChannel(ch_name=line.ch_name, var_name=line.name))
        in_chs_head.sort(key=operator.attrgetter("ch_name"))
        # Then, get all the output channels of the head block
        out_chs_head = []
        for lines in head_block.src_lines:
            for line in lines:
                if line.dest not in all_blocks:
                    out_chs_head.append(hp.OutputChannel(ch_name=line.ch_name, expr=AVar(line.name)))
        out_chs_head.sort(key=operator.attrgetter("ch_name"))
        # Finally, get the HCSP process of the head block
        head_processes = in_chs_head + [get_block_hp(head_block)] + out_chs_head
        # After translation, the names of the dest_lines of the head block should be changed back
        for line in head_block.dest_lines:
            if line.name.startswith("delay_"):
                line.name = line.name[6:]
        head_process = head_processes[0] if len(head_processes) == 1 else hp.Sequence(*head_processes)
        cond_time = RelExpr("==", OpExpr("%", AVar("t"), AConst(head_block.st)), AConst(0))
        main_processes.append(hp.Condition(cond_time, head_process))

        # Delete the head block from the loop
        del block_dict[head_block.name]

        # Continue the topological sorting after deleting the head block from the loop
        topo_sort()

    # Get block dictionary
    block_dict = {block.name: block for block in diagram}
    all_blocks = list(block_dict.keys())

    # The case that there are only constant blocks in block_dict
    # if all(block.type == "constant" for block in block_dict.values()):
    #     assert len(block_dict) == 1
    #     _, block = block_dict.popitem()
    #     line = block.src_lines[0][0]
    #     return hp.Condition(cond=RelExpr("==", ModExpr(AVar("t"), AConst(block.st)), AConst(0)),
    #                         hp=hp.OutputChannel(ch_name=line.ch_name, expr=AConst(block.value)))

    # Get initial processes
    init_hp = hp.Assign("t", AConst(0))
    for block in block_dict.values():
        if block.type == "constant":
            var_map = block.get_var_map()
            assert len(var_map) == 1
            (out_var, [(_, expr)]) = var_map.popitem()
            assert isinstance(expr, AConst)
            init_hp = hp.Sequence(init_hp, hp.Assign(out_var, expr))
        elif block.type == "unit_delay":
            init_hp = hp.Sequence(init_hp,
                                  hp.Assign("delay_"+block.src_lines[0][0].name,
                                            hp.AConst(block.init_value)))

    # Delete Constant blocks
    constant_block_names = [name for name, block in block_dict.items() if block.type == "constant"]
    for name in constant_block_names:
        del block_dict[name]

    # Get diagram sample time and the wait process
    diagram_st = get_gcd([block.st for block in block_dict.values()]) if len(block_dict) >0 else 1
    wait_st = hp.Sequence(hp.Wait(AConst(diagram_st)),
                          hp.Assign("t", OpExpr("+", AVar("t"), AConst(diagram_st))))

    # Get main processes
    main_processes = []
    topo_sort()

    # Get loop body of the result process
    main_processes.append(wait_st)
    result_hp = hp.Sequence(init_hp, hp.Loop(hp.Sequence(*main_processes)))
    return result_hp


def new_translate_discrete(diagram, chart_parameters):
    """Obtain procedures for the discrete part of the diagram."""
    assert isinstance(diagram, list)  # diagram is a list of blocks
    sample_time = get_gcd([block.st for block in diagram
                          if isinstance(block.st, (int, Decimal)) and block.st > 0])
    block_dict = {block.name: block for block in diagram}

    # Record initializations
    init_hps = list()

    # Storing procedures generated when translating
    procedures = list()

    # Delete in_ and out_ports from block_dict
    port_names = [name for name, block in block_dict.items() if block.type in ("in_port", "out_port")]
    for port_name in port_names:
        del block_dict[port_name]

    # Translate Stateflow charts
    charts = [block for block in block_dict.values() if block.type == "stateflow"]

    # First, for the charts triggered by function calls, obtain mapping
    # from events to corresponding charts.
    fun_call_map = dict()
    for chart in charts:
        for trigger_type, event in chart.input_events:
            if trigger_type == 'function-call':
                fun_call_map[event] = chart.name
    for chart in charts:
        converter = sf_convert.SFConvert(
            chart, chart_parameters=chart_parameters[chart.name], translate_io=False,
            fun_call_map=fun_call_map)
        init_hps.append(hcsp.Var(converter.init_name(chart.name)))
        chart.exec_name = converter.exec_name(charts[0].name)
        _procedures = converter.get_procs()
        for _name, _hp in _procedures.items():
            procedures.append(hcsp.Procedure(_name, _hp))

    # Generate init_hp
    for block in block_dict.values():
        if block.type == "constant":
            out_var = block.src_lines[0][0].name
            init_hps.append(hp.Assign(var_name=out_var, expr=AConst(block.value)))
        elif block.type == "unit_delay":
            init_hps.append(hp.Assign(var_name=block.name+"_state", expr=AConst(block.init_value)))
        elif block.type == "discrete_PID_controller":
            init_hps.extend(block.get_init_hps())
        elif block.type == "triggered_subsystem":
            init_hps.extend(block.get_init_hps())
            procedures.extend(block.get_procedures())
        elif block.type == "signalBuilder":
            init_hps.append(block.get_init_hp())
        elif block.type == "stateflow":
            init_hps.extend(block.get_init_hps())
        elif block.type == "DiscretePulseGenerator":
            init_hps.append(block.get_init_hp())

    # Delete Constant blocks
    block_names = [name for name, block in block_dict.items() if block.type == "constant"]
    for name in block_names:
        del block_dict[name]

    # Get a topological sort of the blocks
    # First, move all the Unit_Delay blocks from block_dict to sorted_blocks
    sorted_blocks = [block for block in block_dict.values() if block.type == "unit_delay"]
    for block in sorted_blocks:
        del block_dict[block.name]

    # Get a topological sort of the remaining blocks
    while block_dict:
        head_block_names = []
        for name, block in block_dict.items():
            src_blocks = block.get_src_blocks()
            if src_blocks.isdisjoint(set(block_dict.keys())):
                sorted_blocks.append(block)
                head_block_names.append(name)
        assert head_block_names
        for name in head_block_names:
            del block_dict[name]

    # Get the OUTPUT of each block in sorted_blocks
    output_hps = []
    update_hps = []
    for block in sorted_blocks:
        if block.st == sample_time:
            # Block runs at every sample time
            output_hps.append(block.get_output_hp())
        elif block.st < 0:
            # Block may be triggered, add the triggering code
            output_hps.append(block.get_output_hp())
        else:
            assert block.st % sample_time == 0
            period = block.st // sample_time
            output_hps.append(hp.ITE([(
                RelExpr("==", OpExpr("%", AVar("_tick"), AConst(period)), AConst(0)),
                block.get_output_hp())]))

        if block.type == "unit_delay":
            if block.st == sample_time:
                update_hps.append(block.get_update_hp())
            else:
                assert block.st % sample_time == 0
                period = block.st // sample_time
                update_hps.append(hp.ITE([(
                    RelExpr("==", OpExpr("%", AVar("_tick"), AConst(period)), AConst(0)),
                    block.get_update_hp())]))
    return init_hps, procedures, output_hps, update_hps, sample_time


def new_translate_continuous(diagram):
    """Translate continuous diagram.

    diagram : SL_Diagram

    Returns five-tuple:
    init_hps: initialization processes.
    equations: equations of the ODE.
    constraints: constraint of the ODE.
    trig_procs: triggered procedures.
    procedures: list of procedures from triggered subsystems.

    """
    # tt is the LOCAL evolution time of continuous process
    init_hps = [hp.Assign(var_name="tt", expr=AConst(0))]
    equations = [("tt", AConst(1))]  # tt_dot = 1
    constraints = []
    trig_procs = []
    procedures = []

    # Dictionary of variable substitutions. There should be no loops
    # in substitution.
    var_subst = dict()

    for block in diagram:
        if block.type in ('add', 'product', 'bias', 'gain', 'constant', 'square',
                          'sqrt', 'switch', 'sine'):
            var_subst.update(block.get_var_subst())
        elif block.type == "integrator":
            in_var = block.dest_lines[0].name
            out_var = block.src_lines[0][0].name
            init_hps.append(hp.Assign(out_var, AConst(block.init_value)))
            equations.append((out_var, AVar(in_var)))
            if block.enable != true_expr:
                constraints.append(block.enable)
        elif block.type == "triggered_subsystem":
            init_hps.extend(block.get_init_hps())
            init_hps.append(hp.Assign(block.triggered, AConst(0)))
            trig_cond = block.get_continuous_triggered_condition()
            trig_procs.append((trig_cond, hp.Var(block.name)))
            constraints.append(neg_expr(trig_cond))
            procedures.extend(block.get_procedures())
        elif block.type in ('scope'):  # ignore
            pass
        else:
            raise NotImplementedError('Unrecognized continuous block: %s' % block.type)

    for i in range(len(equations)):
        var, e = equations[i]
        equations[i] = (var, subst_all(e, var_subst))

    return init_hps, equations, var_subst, constraints, trig_procs, procedures


def new_get_hcsp(discrete_diagram, continuous_diagram, chart_parameters, outputs=()):
    dis_init_hps, dis_procedures, output_hps, update_hps, sample_time = \
        new_translate_discrete(discrete_diagram, chart_parameters)
    con_init_hps, equations, var_subst, constraints, trig_procs, con_procedures = \
        new_translate_continuous(continuous_diagram)

    # Initialization
    init_hps = [hp.Assign("t", AConst(0)), hp.Assign("_tick", AConst(0))] + dis_init_hps + con_init_hps
    init_hp = init_hps[0] if len(init_hps) == 1 else hp.Sequence(*init_hps)

    ### Discrete process ###
    discrete_hps = output_hps + update_hps
    discrete_hp = hp.seq(discrete_hps)
    discrete_hp = hp.subst_comm_all(discrete_hp, var_subst)

    ### Continuous process ###

    # If sample_time = 0, the entire diagram is continuous. Arbitrarily
    # choose 1 as sample time
    if sample_time == 0:
        sample_time = 1

    # Add tt < sample_time to the constraint
    time_constraint = RelExpr("<", AVar("tt"), AConst(sample_time))
    constraints.append(time_constraint)

    # Form ODE
    continuous_hp = hp.ODE(eqs=equations, constraint=conj(*constraints))
    names_triggered = None
    if trig_procs:
        names_triggered = list()
        for _, sys_name in trig_procs:
            name_triggered = sys_name.name + "_triggered"
            names_triggered.append(
                hp.ITE([(RelExpr(">", AVar(name_triggered), AConst(0)),
                        hp.Assign(var_name=name_triggered,
                                          expr=OpExpr("-", AVar(name_triggered), AConst(1))))]))
        names_triggered = hp.seq(names_triggered)
        trig_proc = list()
        for cond, sys_name in trig_procs:
            set_triggered = hp.ITE(if_hps=[(RelExpr("<", AVar("tt"), AConst(sample_time)),
                                            hp.Assign(var_name=sys_name.name+"_triggered", expr=AConst(1)))],
                                   else_hp=hp.Assign(var_name=sys_name.name+"_triggered", expr=AConst(2))
                                   )
            trig_proc.append(hp.ITE([(cond, hp.Sequence(sys_name, set_triggered))]))
        trig_proc = hp.Sequence(*trig_proc) if len(trig_proc) >= 2 else trig_proc[0]
        continuous_hp = hp.Loop(hp=hp.Sequence(continuous_hp, trig_proc), constraint=time_constraint)

    # Update t := t + tt
    update_t = hp.Assign("t", OpExpr("+", AVar("t"), AVar("tt")))

    # Update tick := tick + 1
    update_tick = hp.Assign("_tick", OpExpr("+", AVar("_tick"), AConst(1)))

    # Reset tt := 0
    reset_tt = hp.Assign(var_name="tt", expr=AConst(0))

    if names_triggered:
        continuous_hp = hp.Sequence(names_triggered, continuous_hp, update_t, update_tick, reset_tt)
    else:
        continuous_hp = hp.Sequence(continuous_hp, update_t, update_tick, reset_tt)

    # Main process
    main_hp = hp.Sequence(init_hp, hp.Loop(hp.Sequence(discrete_hp, continuous_hp)))

    # Get procedures
    procedures = dis_procedures + con_procedures

    dict_procs = {proc.name: proc.hp for proc in procedures}
    # dict_procs, main_hp = full_optimize_module(dict_procs, main_hp)
    procedures = [hp.Procedure(name=proc_name, hp=proc_hp) for proc_name, proc_hp in dict_procs.items()]
    return HCSPModule(name="P", code=main_hp, procedures=procedures, outputs=outputs)


def get_hcsp(dis_subdiag_with_chs, con_subdiag_with_chs, sf_charts, buffers,
             discretePulseGenerator, muxs, dataStoreMemorys, dataStoreReads, clocks, model_name="P"):
    """Obtain HCSP from a list of disjoint diagrams.
    
    The arguments are:
    dis_subdiag_with_chs - list of discrete diagrams
    con_subdiag_with_chs - list of continuous diagrams
    sf_charts - list of Stateflow charts
    unit_delays - list of unit delay blocks
    buffers - list of buffers

    """
    processes = hp.HCSPProcess()
    main_processes = []
    # triggered_process = []
    # Compute the discrete processes from discrete subdiagrams
    num = 0
    for data in dataStoreMemorys:
        name = "DSM" + str(num)
        processes.add(name, data.get_hcsp())
        main_processes.append(hp.Var(name))
        num += 1
    num = 0
    for diag in dis_subdiag_with_chs:
        name = "PD" + str(num)
        # discrete_process = translate_discrete(diag)
        discrete_process = translate_discrete(diag)
        processes.add(name, discrete_process)
        main_processes.append(hp.Var(name))
        num += 1

    # Compute the continuous processes from continuous subdiagrams
    num = 0
    for diag in con_subdiag_with_chs:
        name = "PC" + str(num)
        continuous_process = translate_continuous(diag)
        processes.add(name, continuous_process)
        main_processes.append(hp.Var(name))
        num += 1
    num=0
    for block in discretePulseGenerator:
        for line in block.src_lines:
            for branch in line:
                if isinstance(branch.dest_block,Mux):
                    plus_hcsp = block.get_hcsp()
                    name = "DPG" + str(num)
                    processes.add(name, plus_hcsp)
                    main_processes.append(hp.Var(name))
                else:
                    plus_hcsp = block.get_hcsp1()
                    name = "DPG" + str(num)
                    processes.add(name, plus_hcsp)
                    main_processes.append(hp.Var(name))
        num += 1
    num1=0
    for block in muxs:
        name="Mux"+str(num1)
        processes.add(name, block.get_hp())
        main_processes.append(hp.Var(name))
        num1+=1

    # Compute the stateflow processes
    for chart in sf_charts:
        chart.sf_charts = sf_charts

    # Computer the buffer processes
    for buffer in buffers:
        process_name = buffer.name.replace(" ", "_")
        processes.add(process_name, buffer.get_hcsp())
        main_processes.append(hp.Var(process_name))

    # Get main paralell processes
    assert len(main_processes) >= 1 and len(main_processes) == len(processes.hps)
    if len(main_processes) == 1:
        processes.hps = [(model_name, processes.hps[0][1])]
    else:
            main_process = hp.Parallel(*main_processes)
            processes.insert(n=0, name=model_name, hp=main_process)

    return processes


if __name__ == "__main__":
    import sys
    if len(sys.argv) != 2:
        print("Usage: python sf_convert.py filename")
    else:
        filename = sys.argv[1]
        diagram = SL_Diagram(filename)
        diagram.parse_xml()
        diagram.delete_subsystems()
        diagram.add_line_name()
        diagram.comp_inher_st()
        diagram.inherit_to_continuous()
        diagram.separate_diagram()

        # Convert to HCSP
        result_hp = new_get_hcsp(
            diagram.discrete_blocks, diagram.continuous_blocks,
            diagram.chart_parameters, diagram.outputs)

        # Optimize module
        hp = result_hp.code
        procs = dict((proc.name, proc.hp) for proc in result_hp.procedures)
        procs, hp = full_optimize_module(procs, hp)
        result_hp.procedures = [hcsp.Procedure(name, hp) for name, hp in procs.items()]
        result_hp.code = hp

        print(result_hp.export())



"""Events in Stateflow."""

class SF_Event:
	def __init__(self, name="", trigger=None, scope=None, port=None):
		self.name = name
		self.trigger = trigger
		self.scope = scope
		self.port = port

	def __str__(self):
		return "{ name:%s ; trigger:%s ; scope:%s,port:%s}" % \
			(self.name, self.trigger,self.scope,self.port)



# Unit tests for expressions

import unittest

from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp.expr import true_expr, AVar, AConst
from ss2hcsp.hcsp.parser import expr_parser, expr_parser, hp_parser
from ss2hcsp.hcsp import hcsp


class ExprTest(unittest.TestCase):
    def testGetRange(self):
        self.assertEqual(expr.get_range(0.1, 0.4), [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(expr.get_range(0.03, 0.41), [0.03, 0.1, 0.2, 0.3, 0.4, 0.41])
        self.assertEqual(expr.get_range(0.1, 0.41), [0.1, 0.2, 0.3, 0.4, 0.41])
        self.assertEqual(expr.get_range(0.03, 0.4), [0.03, 0.1, 0.2, 0.3, 0.4])

    def testExprParser(self):
        test_data = [
            "a + 1",
            "a * b",
            "a - b",
            "min(a,b)",
            "a * b + c",
            "a * (b + c)",
            "a + b - c",
            "a + (b - c)",
            "[]",
            "\"a\"",
            "[b,0]",
            "a[0]",
            "-x",
            "-x + y",
            "(a - b) % m",
            "(a > 0 ? a : -a)",
            "(a > 0 ? (b > 0 ? b : -b) : -a)"
        ]
        
        for s in test_data:
            expr = expr_parser.parse(s)
            self.assertEqual(str(expr), s)

    def testExprParser(self):
        test_data = [
            "a < 1",
            "a == 1 && true",
            "a == 1 && b == 2 && c == 3",
            "(a == 1 && b == 2) && c == 3",
            "a == 1 && b == 2 || c == 3",
            "a == 1 || b == 2 && c == 3",
            "(a == 1 || b == 2) && c == 3",
            "(a == 1 || b == 2 || c == 3) && d == 4",
        ]

        for s in test_data:
            expr = expr_parser.parse(s)
            self.assertEqual(str(expr), s)

    def assertConditionalInst(self, test_data, res):
        inst = expr.Conditional_Inst()
        for var_name, cond_inst_str in test_data:
            cond_inst = [(expr_parser.parse(cond), expr_parser.parse(inst))
                         for cond, inst in cond_inst_str]
            inst.add(var_name, cond_inst)

        res_inst = dict()
        for var_name, cond_inst_str in res:
            cond_inst = [(expr_parser.parse(cond), expr_parser.parse(inst))
                         for cond, inst in cond_inst_str]
            res_inst[var_name] = cond_inst
        self.assertEqual(inst.data, res_inst)

    def testConditionalInst1(self):        
        test_data = [
            ("x", [("true", "a + 1")]),
            ("y", [("a >= 1", "x"), ("a < 1", "x - 1")]), 
        ]

        res = [
            ("x", [("true", "a + 1")]),
            ("y", [("a >= 1", "a + 1"), ("a < 1", "(a + 1) - 1")])
        ]

        self.assertConditionalInst(test_data, res)

    def testConditionalInst2(self):
        test_data = [
            ("w", [("in >= 5", "a"), ("in < 5", "b")]),
            ("z", [("w >= 10", "x"), ("w < 10", "y")]),
        ]

        res = [
            ("w", [("in >= 5", "a"), ("in < 5", "b")]),
            ("z", [("a >= 10 && in >= 5 || b >= 10 && in < 5", "x"),
                   ("a < 10 && in >= 5 || b < 10 && in < 5", "y")]),
        ]

        self.assertConditionalInst(test_data, res)

    def testParseHCSP(self):
        test_data = [
            "x1 := 3;",
            "x1 := 3; x2 := 5;",
            "x1 := 3; x2 := 5; skip;",
            "{skip;}*",
            "x1?x1;",
            "x1!x2;",
            "{x_dot = x + 1, y_dot = y + 1 & x < 3} |> [] (x?x --> skip;, y!y --> skip;)",
            "x?x --> skip; $ y!y --> x := 2;",
            "@A; @B; || @C;",
            "@A; @B; || @C; @D;",
            "if (x == 0) { @A; }",
            "if (x == 0) { @A; @B; }",
            "if (x == 0) { @A; @B; }",
            "x?; y?; || z?;",
            "if (x == 1) { x := 0; } else { x := 1; }",
            "if (x == 0) { x := 1; } else if (x == 1) { x := 0; } else { skip; }",
            'E := "x";',
            'EL := ["x"];',
            'EL := ["x","y"];',
            'NL := [1];',
            'NL := [1,2.1];',
            "{x := x + 1;}*(x < 3)",
            "(a, b) := xs;",
            "x := -x;",
            "x := -x + y;",
            "assert(x == 2);",
            "assert(x == 2,\"message\");",
            "log(\"start\");",
            "pt.x := pt.y + 1;",
            "ch!pt.y;",
            "pt := {x:1,y:2};",
            "ch!a[1][2];",
            "ch!pt.x[2];",
            "a[1][2]!x;",
            "a[1][2]?pt.x[1];",
            "@X;",
            "x := (a - b) % m;",
            "x := x + 1; ++ y := y + 1;",
        ]

        for s in test_data:
            hp = hp_parser.parse(s)
            self.assertEqual(str(hp), s)


if __name__ == "__main__":
    unittest.main()



"""Optimization of HCSP code."""

from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import hcsp
from ss2hcsp.hcsp import simulator


def is_atomic(hp):
    """Whether hp is an atomic program"""
    return hp.type in ('skip', 'wait', 'assign', 'assert', 'test', 'log',
                       'input_channel', 'output_channel', 'ode')

def simplify_expr(e):
    assert isinstance(e, expr.Expr)
    if not e.get_vars():
        b = simulator.eval_expr(e, dict())
        assert isinstance(b, bool)
        return expr.true_expr if b else expr.false_expr
    elif isinstance(e, expr.LogicExpr):
        if e.op == '&&':
            return expr.conj(*(simplify_expr(arg) for arg in e.exprs))
        elif e.op == '||':
            return expr.disj(*(simplify_expr(arg) for arg in e.exprs))
        elif e.op == '->':
            return expr.imp(simplify_expr(e.exprs[0]), simplify_expr(e.exprs[1]))
        else:
            return e
    else:
        return e

def simplify(hp):
    """Perform immediate simplifications to HCSP process. This includes:
    
    * Remove extraneous skip from processes.
    * Simplify if true then P else Q to P.
    
    """
    if is_atomic(hp) or hp.type == 'var':
        return hp
    elif hp.type == 'sequence':
        return hcsp.seq([simplify(sub_hp) for sub_hp in hp.hps])
    elif hp.type == 'loop':
        return hcsp.Loop(simplify(hp.hp), constraint=hp.constraint)
    elif hp.type == 'ode':
        return hcsp.ODE(hp.eqs, simplify(hp.constraint), out_hp=simplify(hp.out_hp))
    elif hp.type == 'ode_comm':
        return hcsp.ODE_Comm(hp.eqs, hp.constraint,
                             [(io, simplify(comm_hp)) for io, comm_hp in hp.io_comms])
    elif hp.type == 'select_comm':
        return hcsp.SelectComm(*((io, simplify(comm_hp)) for io, comm_hp in hp.io_comms))
    elif hp.type == 'ite':
        new_if_hps = []
        for cond, if_hp in hp.if_hps:
            simp_cond = simplify_expr(cond)
            simp_if_hp = simplify(if_hp)
            if simp_if_hp.type == 'skip' or simp_cond == expr.false_expr:
                continue
            elif simp_cond == expr.true_expr:
                if len(new_if_hps) > 0:
                    return hcsp.ite(new_if_hps, simp_if_hp)
                else:
                    return simp_if_hp
            elif simp_if_hp.type == 'ite' and simp_if_hp.if_hps[0][0] == simp_cond:
                simp_if_hp = simp_if_hp.if_hps[0][1]
            elif simp_if_hp.type == 'sequence' and simp_if_hp.hps[0].type == 'ite' \
                and simp_if_hp.hps[0].if_hps[0][0] == simp_cond:
                simp_if_hp = hcsp.Sequence(simp_if_hp.hps[0].if_hps[0][1], *simp_if_hp.hps[1:])
            new_if_hps.append((simp_cond, simp_if_hp))
        if hp.else_hp is None:
            new_else_hp = None
        else:
            new_else_hp = simplify(hp.else_hp)
        return hcsp.ite(new_if_hps, new_else_hp)
    else:
        raise NotImplementedError

def get_read_vars_lname(lname):
    """Return the set of variables that are read in an expression to be
    assigned to. This does *not* include the variable that is assigned.
    
    Example: if lname = a[i+j], then the result is {i, j}.
    
    """
    if lname is None:
        return set()
    elif isinstance(lname, expr.AVar):
        return set()
    elif isinstance(lname, expr.ArrayIdxExpr):
        return lname.expr2.get_vars().union(get_read_vars_lname(lname.expr1))
    elif isinstance(lname, expr.FieldNameExpr):
        return get_read_vars_lname(lname.expr)
    else:
        raise NotImplementedError

def replace_read_vars_lname(lname, inst):
    """Replace read variables in an expression to be assigned to. This
    does *not* replace the variable that is assigned.
    
    """
    if lname is None:
        return None
    elif isinstance(lname, expr.AVar):
        return lname
    elif isinstance(lname, expr.ArrayIdxExpr):
        return expr.ArrayIdxExpr(replace_read_vars_lname(lname.expr1, inst), lname.expr2.subst(inst))
    elif isinstance(lname, expr.FieldNameExpr):
        return expr.FieldNameExpr(replace_read_vars_lname(lname.expr, inst), lname.field)
    else:
        raise NotImplementedError

def get_read_vars(hp):
    """Obtain set of variables read by the program."""
    if hp.type in ('skip', 'var'):
        return set()
    elif hp.type == 'wait':
        return hp.delay.get_vars()
    elif hp.type == 'assign':
        return get_read_vars_lname(hp.var_name).union(hp.expr.get_vars())
    elif hp.type in ('assert', 'test', 'log'):
        return hp.get_vars()
    elif hp.type == 'input_channel':
        return get_read_vars_lname(hp.var_name)
    elif hp.type == 'output_channel':
        return hp.get_vars()
    elif hp.type == 'ite':
        return set().union(*(if_cond.get_vars() for if_cond, _ in hp.if_hps))
    elif hp.type == 'ode':
        return hp.constraint.get_vars().union(*(eq.get_vars().union({v}) for v, eq in hp.eqs))
    else:
        raise NotImplementedError

def replace_read_vars(hp, inst):
    """Replace the set of read variables in the program. Return the
    new program.
    
    """
    if hp.type == 'skip':
        return hp
    elif hp.type == 'wait':
        return hcsp.Wait(hp.delay.subst(inst))
    elif hp.type == 'assign':
        return hcsp.Assign(replace_read_vars_lname(hp.var_name, inst), hp.expr.subst(inst))
    elif hp.type == 'assert':
        return hcsp.Assert(hp.bexpr.subst(inst), [msg.subst(inst) for msg in hp.msgs])
    elif hp.type == 'test':
        return hcsp.Test(hp.bexpr.subst(inst), [msg.subst(inst) for msg in hp.msgs])
    elif hp.type == 'log':
        return hcsp.Log(hp.pattern.subst(inst), exprs=[e.subst(inst) for e in hp.exprs])
    elif hp.type == 'input_channel':
        return hcsp.InputChannel(hp.ch_name, replace_read_vars_lname(hp.var_name, inst))
    elif hp.type == 'output_channel':
        if len(hp.ch_name.args) > 0:
            raise NotImplementedError
        if hp.expr is None:
            return hp
        else:
            return hcsp.OutputChannel(hp.ch_name, hp.expr.subst(inst))
    elif hp.type == 'ode':
        return hcsp.ODE([(var, eq.subst(inst)) for var, eq in hp.eqs], hp.constraint.subst(inst))
    else:
        raise NotImplementedError

def get_write_vars_lname(lname):
    """Given lname of an assignment, return the set of variables written."""
    if lname is None:
        return {}
    elif isinstance(lname, expr.AVar):
        return {lname.name}
    elif isinstance(lname, expr.ArrayIdxExpr):
        return get_write_vars_lname(lname.expr1)
    elif isinstance(lname, expr.FieldNameExpr):
        return get_write_vars_lname(lname.expr)
    else:
        raise NotImplementedError

def get_write_vars(hp):
    """Returns the set of written variables for an HCSP program."""
    if hp.type == 'assign':
        return get_write_vars_lname(hp.var_name)
    elif hp.type == 'input_channel':
        return get_write_vars_lname(hp.var_name)
    elif hp.type == 'ode':
        res = set()
        for var, _ in hp.eqs:
            res = res.union({var})
        return res
    else:
        return set()

def get_write_vars_pre(hp):
    """Returns the set of written variables for an HCSP program, that should
    be considered at the before label of the program.
    
    """
    if hp.type == 'ode':
        res = set()
        for var, _ in hp.eqs:
            res = res.union({var})
        return res
    else:
        return set()

def get_write_vars_val(hp, var):
    """Returns the value of the written variables in an HCSP program. If the
    written value is uncertain, return None.
    
    """
    if hp.type == 'assign':
        if isinstance(hp.var_name, expr.AVar) and isinstance(hp.expr, expr.AConst):
            assert hp.var_name == expr.AVar(var)
            return hp.expr
        else:
            return None
    else:
        return None

def targeted_replace(hp, repls):
    """Make the given list of replacements."""
    def rec(hp, cur_pos):
        if is_atomic(hp):
            if cur_pos in repls:
                return replace_read_vars(hp, repls[cur_pos])
            else:
                return hp
        elif hp.type == 'var':
            return hp
        elif hp.type == 'sequence':
            new_hps = []
            for i, sub_hp in enumerate(hp.hps):
                new_hps.append(rec(sub_hp, cur_pos + (i,)))
            return hcsp.Sequence(*new_hps)
        elif hp.type == 'ite':
            new_if_hps = []
            for i, (cond, if_hp) in enumerate(hp.if_hps):
                new_if_hp = rec(if_hp, cur_pos + (i,))
                if cur_pos in repls:
                    new_if_hps.append((cond.subst(repls[cur_pos]), new_if_hp))
                else:
                    new_if_hps.append((cond, new_if_hp))
            if hp.else_hp is None:
                new_else_hp = None
            else:
                new_else_hp = rec(hp.else_hp, cur_pos + (len(hp.if_hps),))
            return hcsp.ITE(new_if_hps, new_else_hp)
        elif hp.type == 'loop':
            return hcsp.Loop(rec(hp.hp, cur_pos), constraint=hp.constraint)
        elif hp.type == 'select_comm':
            new_io_comms = []
            for i, (comm_hp, out_hp) in enumerate(hp.io_comms):
                new_out_hp = rec(out_hp, cur_pos + (i,))
                new_io_comms.append((comm_hp, new_out_hp))
            return hcsp.SelectComm(*new_io_comms)
        elif hp.type == 'ode_comm':
            new_io_comms = []
            for i, (comm_hp, out_hp) in enumerate(hp.io_comms):
                new_out_hp = rec(out_hp, cur_pos + (i,))
                new_io_comms.append((comm_hp, new_out_hp))
            return hcsp.ODE_Comm(hp.eqs, hp.constraint, new_io_comms)
        else:
            raise NotImplementedError
    return rec(hp, tuple())

def targeted_remove(hp, remove_locs):
    """Remove code at the given location. For simplicity, simply change
    them to Skip.

    """
    def rec(hp, cur_pos):
        if is_atomic(hp):
            if cur_pos in remove_locs:
                return hcsp.Skip()
            else:
                return hp
        elif hp.type == 'var':
            return hp
        elif hp.type == 'sequence':
            new_hps = []
            for i, sub_hp in enumerate(hp.hps):
                new_hps.append(rec(sub_hp, cur_pos + (i,)))
            return hcsp.Sequence(*new_hps)
        elif hp.type == 'ite':
            new_if_hps = []
            for i, (cond, if_hp) in enumerate(hp.if_hps):
                new_if_hp = rec(if_hp, cur_pos + (i,))
                new_if_hps.append((cond, new_if_hp))
            if hp.else_hp is None:
                new_else_hp = None
            else:
                new_else_hp = rec(hp.else_hp, cur_pos + (len(hp.if_hps),))
            return hcsp.ITE(new_if_hps, new_else_hp)
        elif hp.type == 'loop':
            return hcsp.Loop(rec(hp.hp, cur_pos), constraint=hp.constraint)
        elif hp.type == 'select_comm':
            new_io_comms = []
            for i, (comm_hp, out_hp) in enumerate(hp.io_comms):
                new_out_hp = rec(out_hp, cur_pos + (i,))
                new_io_comms.append((comm_hp, new_out_hp))
            return hcsp.SelectComm(*new_io_comms)
        elif hp.type == 'ode_comm':
            new_io_comms = []
            for i, (comm_hp, out_hp) in enumerate(hp.io_comms):
                new_out_hp = rec(out_hp, cur_pos + (i,))
                new_io_comms.append((comm_hp, new_out_hp))
            return hcsp.ODE_Comm(hp.eqs, hp.constraint, new_io_comms)
        else:
            raise NotImplementedError
    return rec(hp, tuple())


class LocationInfo:
    """Information for each location obtained during static analysis."""
    def __init__(self, loc, sub_hp):
        self.loc = loc
        self.sub_hp = sub_hp

        # The following are computed at initialization
        self.edges = []
        self.rev_edges = []
        self.exits = []

        # Reaching definitions
        self.reach_before = set()
        self.reach_after = set()

        # Live variables
        self.live_before = set()
        self.live_after = set()

    def __str__(self):
        lines = []
        lines.append("Location %s" % str(self.loc))
        if is_atomic(self.sub_hp):
            lines.append("  Code: %s" % str(self.sub_hp))
        lines.append("  Edges: %s" % (', '.join(str(edge) for edge in self.edges)))
        lines.append("  Edges (rev): %s" % (', '.join(str(edge) for edge in self.rev_edges)))
        return '\n'.join(lines)

    def __repr__(self):
        return "LocationInfo(%s)" % (repr(self.loc), repr(self.sub_hp))


class HCSPAnalysis:
    def __init__(self, hp, *, local_vars=None):
        self.hp = hp
        if local_vars is None:
            local_vars = set()
        self.local_vars = local_vars
        self.infos = dict()

        self.all_vars = set()

        self.init_infos()
        self.init_all_vars()
    
    def add_edge(self, loc1, loc2):
        self.infos[loc1].edges.append(loc2)
        self.infos[loc2].rev_edges.append(loc1)

    def init_infos(self):
        """Initialize location infos."""
        def rec(hp, cur_pos):
            # Note location names are repeated for loops
            if cur_pos not in self.infos:
                self.infos[cur_pos] = LocationInfo(cur_pos, hp)

            if is_atomic(hp) or hp.type == 'var':
                self.infos[cur_pos].exits.append(cur_pos)

            elif hp.type == 'sequence':
                for i, sub_hp in enumerate(hp.hps):
                    rec(sub_hp, cur_pos + (i,))
                
                # Entry into first command
                self.add_edge(cur_pos, cur_pos + (0,))

                # Entry into remaining commands
                for i in range(1, len(hp.hps)):
                    for exit_loc in self.infos[cur_pos + (i-1,)].exits:
                        self.add_edge(exit_loc, cur_pos + (i,))
    
                # Overall exit
                self.infos[cur_pos].exits = self.infos[cur_pos + (len(hp.hps)-1,)].exits

            elif hp.type == 'ite':
                for i, (cond, if_hp) in enumerate(hp.if_hps):
                    rec(if_hp, cur_pos + (i,))
                if hp.else_hp is not None:
                    rec(hp.else_hp, cur_pos + (len(hp.if_hps),))

                if hp.else_hp is None:
                    self.infos[cur_pos].exits.append(cur_pos)

                # Possible entry into each branch
                for i in range(len(hp.if_hps) + (0 if hp.else_hp is None else 1)):
                    self.add_edge(cur_pos, cur_pos + (i,))
                
                # Possible exit from each branch
                for i in range(len(hp.if_hps) + (0 if hp.else_hp is None else 1)):
                    self.infos[cur_pos].exits.extend(self.infos[cur_pos + (i,)].exits)
            
            elif hp.type == 'loop':
                rec(hp.hp, cur_pos)

                # Return from loop
                for exit_loc in self.infos[cur_pos].exits:
                    self.add_edge(exit_loc, cur_pos)

            elif hp.type == 'select_comm' or hp.type == 'ode_comm':
                for i, (comm_hp, out_hp) in enumerate(hp.io_comms):
                    rec(out_hp, cur_pos + (i,))

                # Possible entry into each select
                for i in range(len(hp.io_comms)):
                    self.add_edge(cur_pos, cur_pos + (i,))

                # Possible exit from each branch
                for i in range(len(hp.io_comms)):
                    self.infos[cur_pos].exits.extend(self.infos[cur_pos + (i,)].exits)

            else:
                print(hp.type)
                raise NotImplementedError

        rec(self.hp, tuple())

    def init_all_vars(self):
        """Obtain the list of all variables assigned."""
        for loc, info in self.infos.items():
            self.all_vars = self.all_vars.union(get_write_vars(info.sub_hp))

    def compute_reaching_definition(self):
        """Reaching definition analysis.

        For each program location, compute the set of assignments that
        can possibly reach this location. Each assignment is given by
        the variable assigned and the location where it occurs.

        Note for assignment to array elements, we keep name of the
        array rather than the specific element. Likewise for assignment
        to field names.

        Reference:
        Nielson et al. Principles of Program Analysis, Section 2.1.2.

        """
        # Propagate from reach_before to reach_after
        def propagate_before_after(info, var, val):
            if (var, val) not in info.reach_after and var not in get_write_vars(info.sub_hp):
                info.reach_after.add((var, val))

        # Initial definitions
        for var in self.all_vars:
            self.infos[()].reach_before.add((var, None))
            propagate_before_after(self.infos[()], var, None)

        # Assignments
        for loc, info in self.infos.items():
            for var in get_write_vars(info.sub_hp):
                info.reach_after.add((var, get_write_vars_val(info.sub_hp, var)))
            for var in get_write_vars_pre(info.sub_hp):
                info.reach_before.add((var, get_write_vars_val(info.sub_hp, var)))

        # After procedure calls, anything can happen
        for loc, info in self.infos.items():
            if info.sub_hp.type == 'var':
                for var in self.all_vars:
                    self.infos[loc].reach_after.add((var, None))

        # Use dictionary order as approximation for a good traversal order
        process_order = sorted(self.infos.keys())

        # Propagate reach_after to reach_before of successor
        while True:
            found = False
            for process_loc in process_order:
                info = self.infos[process_loc]
                for var, val in info.reach_after:
                    for next_loc in info.edges:
                        if (var, val) not in self.infos[next_loc].reach_before:
                            found = True
                            self.infos[next_loc].reach_before.add((var, val))
                            propagate_before_after(self.infos[next_loc], var, val)

            if not found:
                break

    def detect_replacement(self):
        """Replacement is possible if a variable occurs in an
        expression and there is only one possible assignment of the
        variable leading to it.
        
        """
        repls = dict()
        for loc, info in self.infos.items():
            if is_atomic(info.sub_hp) or info.sub_hp.type in ('ite'):   
                for var in get_read_vars(info.sub_hp):
                    # Count number of occurrences in var
                    reach_defs = [prev_val for prev_var, prev_val in info.reach_before
                                  if prev_var == var]
                    
                    # Obtain the unique expression assigned to.
                    unique_assign = None
                    for prev_val in reach_defs:
                        if prev_val is None:
                            unique_assign = None
                            break
                        if unique_assign is not None and prev_val != unique_assign:
                            unique_assign = None
                            break
                        unique_assign = prev_val
                    # If there is a unique assigned value which is a constant,
                    # add to replacement list.
                    if not unique_assign:
                        continue
                    if loc not in repls:
                        repls[loc] = dict()
                    repls[loc][var] = unique_assign
        return repls

    def compute_live_variable(self):
        """Live variable analysis.

        Reference:
        Nielson et al. Principles of Program Analysis, Section 2.1.4

        """
        # Propagate from live_after to live_before. Note propagation
        # is forbidden only if the variable is exactly equal to the
        # left hand side of assignment. Assigning to indices or field
        # names does not count.
        def propagate_after_before(info, var):
            if var not in info.live_before and \
                (info.sub_hp.type != "assign" or expr.AVar(var) != info.sub_hp.var_name):
                info.live_before.add(var)

        # All variables are live at the end, except those in local_vars
        for loc in self.infos[()].exits:
            info = self.infos[loc]
            for var in self.all_vars - self.local_vars:
                info.live_after.add(var)
                propagate_after_before(info, var)

        # Live variables for commands
        for loc, info in self.infos.items():
            if is_atomic(info.sub_hp) or info.sub_hp.type in ('ite'):
                for var in get_read_vars(info.sub_hp):
                    self.infos[loc].live_before.add(var)
        
        # Before procedure calls, any variable may be used
        for loc, info in self.infos.items():
            if info.sub_hp.type == 'var':
                for var in self.all_vars - self.local_vars:
                    self.infos[loc].live_before.add(var)

        # Use reverse dictionary order as approximation for good traversal order
        process_order = reversed(sorted(self.infos.keys()))

        # Propagate from reach before to reach after of predecessor
        while True:
            found = False
            for process_loc in process_order:
                info = self.infos[process_loc]
                for var in info.live_before:
                    for prev_loc in info.rev_edges:
                        if var not in self.infos[prev_loc].live_after:
                            found = True
                            self.infos[prev_loc].live_after.add(var)
                            propagate_after_before(self.infos[prev_loc], var)

            if not found:
                break

    def detect_dead_code(self):
        """Dead code are assignments whose variable which is not live afterwards."""
        dead_code = []
        for loc, info in self.infos.items():
            if info.sub_hp.type == 'assign':
                if all(name not in info.live_after for name in get_write_vars(info.sub_hp)):
                    dead_code.append(loc)
        return dead_code


def full_optimize(hp, *, local_vars=None):
    """Full optimization of a single HCSP program.

    hp : HCSP - program to be optimized.
    local_vars : set[str] - list of local variables.

    """
    while True:
        old_hp = hp

        # Replace constants
        analysis = HCSPAnalysis(hp, local_vars=local_vars)
        analysis.compute_reaching_definition()
        repls = analysis.detect_replacement()
        hp = targeted_replace(hp, repls)

        # Deadcode elimination
        analysis = HCSPAnalysis(hp, local_vars=local_vars)
        analysis.compute_live_variable()
        dead_code = analysis.detect_dead_code()
        hp = targeted_remove(hp, dead_code)

        # Simplification
        hp = simplify(hp)

        if hp == old_hp:
            break
    return hp

def full_optimize_module(procs, hp, *, local_vars=None, local_vars_proc=None):
    """Full optimization of a single HCSP program, along with procedures.

    hp : HCSP - program to be optimized.
    procs : dict[str, HCSP] - dictionary of procedures.
    local_vars : set[str] - list of local variables in hp.
    local_vars_proc : set[str] - list of local variables in procs.

    """
    while True:
        old_hp = hp
        old_procs = dict()
        for name, proc in procs.items():
            old_procs[name] = proc

        hp = hcsp.reduce_procedures(hp, procs)
        hp = full_optimize(hp, local_vars=local_vars)
        for name in procs:
            procs[name] = full_optimize(procs[name], local_vars=local_vars_proc)
        
        if hp == old_hp and len(old_procs) == len(procs) and \
            all(procs[name] == old_procs[name] for name in procs):
            break

    return procs, hp